/****************************************************************************
Copyright (c) 2010 cocos2d-x.org

http://www.cocos2d-x.org

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
****************************************************************************/
package org.cocos2d.cocoa;

import org.cocos2d.cocoa.CCGeometry.CCPoint;
import org.cocos2d.cocoa.CCGeometry.CCRect;
import org.cocos2d.cocoa.CCGeometry.CCSize;

public class CCAffineTransform {

	public float a, b, c, d;
	public float tx, ty;

	public static CCAffineTransform CCAffineTransformMake(float a, float b, float c, float d, float tx, float ty) {
		CCAffineTransform t = new CCAffineTransform();
		t.a = a; t.b = b; t.c = c; t.d = d; t.tx = tx; t.ty = ty;
		return t;
	}

	public static CCPoint CCPointApplyAffineTransform(final CCPoint point, final CCAffineTransform t) {
		CCPoint p = new CCPoint();
		p.x = (float)((double)t.a * point.x + (double)t.c * point.y + t.tx);
		p.y = (float)((double)t.b * point.x + (double)t.d * point.y + t.ty);
		return p;
	}

	public static CCSize CCSizeApplyAffineTransform(final CCSize size, final CCAffineTransform t) {
		CCSize s = new CCSize();
		s.width = (float)((double)t.a * size.width + (double)t.c * size.height);
		s.height = (float)((double)t.b * size.width + (double)t.d * size.height);
		return s;
	}

	public static CCAffineTransform CCAffineTransformMakeIdentity() {
		return CCAffineTransformMake(1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f);
	}

	public static final CCAffineTransform CCAffineTransformIdentity = CCAffineTransformMakeIdentity();

	public static CCRect CCRectApplyAffineTransform(final CCRect rect, final CCAffineTransform anAffineTransform) {
		float top    = rect.getMinY();
		float left   = rect.getMinX();
		float right  = rect.getMaxX();
		float bottom = rect.getMaxY();

		CCPoint topLeft = CCPointApplyAffineTransform(CCPoint.Make(left, top), anAffineTransform);
		CCPoint topRight = CCPointApplyAffineTransform(CCPoint.Make(right, top), anAffineTransform);
		CCPoint bottomLeft = CCPointApplyAffineTransform(CCPoint.Make(left, bottom), anAffineTransform);
		CCPoint bottomRight = CCPointApplyAffineTransform(CCPoint.Make(right, bottom), anAffineTransform);

		float minX = Math.min(Math.min(topLeft.x, topRight.x), Math.min(bottomLeft.x, bottomRight.x));
		float maxX = Math.max(Math.max(topLeft.x, topRight.x), Math.max(bottomLeft.x, bottomRight.x));
		float minY = Math.min(Math.min(topLeft.y, topRight.y), Math.min(bottomLeft.y, bottomRight.y));
		float maxY = Math.max(Math.max(topLeft.y, topRight.y), Math.max(bottomLeft.y, bottomRight.y));

		return CCRect.Make(minX, minY, (maxX - minX), (maxY - minY));
	}

	public static CCAffineTransform CCAffineTransformTranslate(final CCAffineTransform t, float tx, float ty) {
		return CCAffineTransformMake(t.a, t.b, t.c, t.d, t.tx + t.a * tx + t.c * ty, t.ty + t.b * tx + t.d * ty);
	}

	public static CCAffineTransform CCAffineTransformRotate(final CCAffineTransform t, float anAngle) {
		float fSin = (float) Math.sin(anAngle);
		float fCos = (float) Math.cos(anAngle);

		return CCAffineTransformMake(	t.a * fCos + t.c * fSin,
										t.b * fCos + t.d * fSin,
										t.c * fCos - t.a * fSin,
										t.d * fCos - t.b * fSin,
										t.tx,
										t.ty);
	}

	public static CCAffineTransform CCAffineTransformScale(final CCAffineTransform t, float sx, float sy) {
		return CCAffineTransformMake(t.a * sx, t.b * sx, t.c * sy, t.d * sy, t.tx, t.ty);
	}

	/* Concatenate `t2' to `t1' and return the result:
     t' = t1 * t2 */
	public static CCAffineTransform CCAffineTransformConcat(final CCAffineTransform t1, final CCAffineTransform t2) {
		return CCAffineTransformMake(	t1.a * t2.a + t1.b * t2.c, t1.a * t2.b + t1.b * t2.d, //a,b
										t1.c * t2.a + t1.d * t2.c, t1.c * t2.b + t1.d * t2.d, //c,d
										t1.tx * t2.a + t1.ty * t2.c + t2.tx,                  //tx
										t1.tx * t2.b + t1.ty * t2.d + t2.ty);                  //ty
	}

	/* Return true if `t1' and `t2' are equal, false otherwise. */
	public static boolean CCAffineTransformEqualToTransform(final CCAffineTransform t1, final CCAffineTransform t2) {
		return (t1.a == t2.a && t1.b == t2.b && t1.c == t2.c && t1.d == t2.d && t1.tx == t2.tx && t1.ty == t2.ty);
	}

	public static CCAffineTransform CCAffineTransformInvert(final CCAffineTransform t) {
		float determinant = 1 / (t.a * t.d - t.b * t.c);

		return CCAffineTransformMake(determinant * t.d, -determinant * t.b, -determinant * t.c, determinant * t.a,
				determinant * (t.c * t.ty - t.d * t.tx), determinant * (t.b * t.tx - t.a * t.ty) );
	}
}
